import logging
import re
from unittest.mock import patch
from uuid import UUID

from workflow_server.server import create_app


def test_version_route():
    flask_app = create_app()

    with flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")
        assert response.status_code == 200
        assert re.match(r"[0-9]*\.[0-9]*\.[0-9]*", response.json["sdk_version"])
        assert response.json["server_version"] == "local"


def test_version_route__with_single_node_file(tmp_path):
    # GIVEN a temporary custom_nodes directory with a test node
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    node_file = custom_nodes_dir / "test_node.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    \"""A test node for processing data.\"""
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find exactly one node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the node should have the correct metadata
    node = nodes[0]
    assert UUID(node["id"])
    assert node["module"] == ["vellum_custom_nodes", "test_node"]
    assert node["name"] == "TestNode"
    assert node["label"] == "Test Node"
    assert "A test node for processing data." in node["description"]
    assert node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "test_node"], "name": "TestNode"},
        "display_data": {
            "comment": {"value": "A test node for processing data."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "7a8b251d-f5ca-462a-b293-071d219460fb",
        "label": "TestNode",
        "outputs": [],
        "ports": [{"id": "a3a0eefd-45d0-4f13-8c58-a836a9f7f9ed", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "a022e36c-9852-4772-9be3-3c6c147fd811", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }


def test_version_route__with_nodes_in_multiple_files(tmp_path):
    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a first node file
    first_node_file = custom_nodes_dir / "first_node.py"
    first_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"""This is Some Node.\"""
"""
    )

    # AND a second node file
    second_node_file = custom_nodes_dir / "second_node.py"
    second_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeOtherNode(BaseNode):
    \"""This is Some Other Node.\"""
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find both nodes
    nodes = response.json["nodes"]
    assert len(nodes) == 2

    # AND the first node should have correct metadata
    some_node = nodes[0]
    assert some_node["label"] == "Some Node"
    assert some_node["description"] == "This is Some Node."
    assert UUID(some_node["id"])
    assert some_node["module"] == ["vellum_custom_nodes", "first_node"]
    assert some_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "first_node"], "name": "SomeNode"},
        "display_data": {"comment": {"value": "This is Some Node."}, "position": {"x": 0.0, "y": 0.0}},
        "id": "1e559c2e-db82-41f0-9ceb-5e89b0c5a0a3",
        "label": "SomeNode",
        "outputs": [],
        "ports": [{"id": "48e39e97-5fd4-471e-b4f2-51d3baf06456", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "e3381fb7-61fc-4c46-ae8e-51fc463b6a59", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }

    # AND the second node should have correct metadata
    some_other_node = nodes[1]
    assert some_other_node["label"] == "Some Other Node"
    assert some_other_node["description"] == "This is Some Other Node."
    assert UUID(some_other_node["id"])
    assert some_other_node["module"] == ["vellum_custom_nodes", "second_node"]
    assert some_other_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "second_node"], "name": "SomeOtherNode"},
        "display_data": {"comment": {"value": "This is Some Other Node."}, "position": {"x": 0.0, "y": 0.0}},
        "id": "7aee541b-b245-4c8a-9137-3e4631d5100c",
        "label": "SomeOtherNode",
        "outputs": [],
        "ports": [{"id": "fb66b46a-d970-4bc9-83ea-70c154c57ddd", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "13fa2714-20b3-4bc3-ab79-621a188e3bfa", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }


def test_version_route__no_custom_nodes_dir(tmp_path):
    # GIVEN a Flask application and an empty temp directory
    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the nodes list should be empty
    assert response.json["nodes"] == []


def test_version_route__with_multiple_nodes_in_file(tmp_path):
    # Create a temporary custom_nodes directory with multiple nodes in one file
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # Create a test node file with multiple nodes
    node_file = custom_nodes_dir / "multiple_nodes.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class ProcessingNode(BaseNode):
    \"""Processes input data.\"""

class TransformationNode(BaseNode):
    \"""Transforms data format.\"""

# This class should not be discovered
class HelperClass:
    pass
"""
    )

    flask_app = create_app()

    # Mock the current working directory to point to our temp directory
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

        assert response.status_code == 200
        nodes = response.json["nodes"]
        assert len(nodes) == 2

        # Nodes should be discovered regardless of their order in the file
        node_names = {node["name"] for node in nodes}
        assert node_names == {"ProcessingNode", "TransformationNode"}

        # Find and assert each node individually
        processing_node = next(node for node in nodes if node["name"] == "ProcessingNode")
        assert processing_node["exec_config"] == {
            "adornments": None,
            "attributes": [],
            "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
            "definition": {"module": ["vellum_custom_nodes", "multiple_nodes"], "name": "ProcessingNode"},
            "display_data": {"comment": {"value": "Processes input data."}, "position": {"x": 0.0, "y": 0.0}},
            "id": "f92c09f0-0434-46cb-829d-a73f801d6343",
            "label": "ProcessingNode",
            "outputs": [],
            "ports": [{"id": "abaa2984-b312-4491-b069-e689759f72c8", "name": "default", "type": "DEFAULT"}],
            "trigger": {"id": "35378c2b-f089-44af-ac37-efe4ea42c817", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "type": "GENERIC",
        }

        transformation_node = next(node for node in nodes if node["name"] == "TransformationNode")
        assert transformation_node["exec_config"] == {
            "adornments": None,
            "attributes": [],
            "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
            "definition": {"module": ["vellum_custom_nodes", "multiple_nodes"], "name": "TransformationNode"},
            "display_data": {"comment": {"value": "Transforms data format."}, "position": {"x": 0.0, "y": 0.0}},
            "id": "09ca32f7-c8f2-4469-97e5-1f288f85127a",
            "label": "TransformationNode",
            "outputs": [],
            "ports": [{"id": "88778117-fbfc-4b44-964b-5a4994aa2f24", "name": "default", "type": "DEFAULT"}],
            "trigger": {"id": "5d096263-7fbf-490a-83b7-e441852b5fb6", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "type": "GENERIC",
        }


def test_version_route__with_invalid_node_file(tmp_path, caplog):
    caplog.set_level(logging.WARNING)

    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a valid node file
    valid_node_file = custom_nodes_dir / "valid_node.py"
    valid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"\"\"This is Some Node.\"\"\"
"""
    )

    # AND an invalid node file with syntax error of missing colon in the class
    invalid_node_file = custom_nodes_dir / "invalid_node.py"
    invalid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode)
    \"\"\"This node has a syntax error.\"\"\"
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find only the valid node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the valid node should have correct metadata
    valid_node = nodes[0]
    assert valid_node["label"] == "Some Node"
    assert valid_node["description"] == "This is Some Node."
    assert UUID(valid_node["id"])
    assert valid_node["module"] == ["vellum_custom_nodes", "valid_node"]
    assert valid_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "valid_node"], "name": "SomeNode"},
        "display_data": {"comment": {"value": "This is Some Node."}, "position": {"x": 0.0, "y": 0.0}},
        "id": "1e559c2e-db82-41f0-9ceb-5e89b0c5a0a3",
        "label": "SomeNode",
        "outputs": [],
        "ports": [{"id": "48e39e97-5fd4-471e-b4f2-51d3baf06456", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "e3381fb7-61fc-4c46-ae8e-51fc463b6a59", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }

    # AND the error should be logged with full traceback
    assert len(caplog.records) > 0
    error_message = caplog.records[0].message
    assert "Failed to load node from module invalid_node" in error_message
    assert "invalid_node.py, line 4" in error_message


def test_version_route__with_attributes(tmp_path):
    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND an addition node file
    node_file = custom_nodes_dir / "addition_node.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode


class MyAdditionNode(BaseNode):
    \"\"\"Custom node that performs simple addition.\"\"\"
    arg1: int
    arg2: int

    class Outputs(BaseNode.Outputs):
        result: int

    def run(self) -> BaseNode.Outputs:
        result = self.arg1 + self.arg2
        return self.Outputs(result=result)
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find the addition node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the node should have the correct metadata
    node = nodes[0]
    assert node["label"] == "My Addition Node"
    assert node["description"] == "Custom node that performs simple addition."
    assert UUID(node["id"])
    assert node["module"] == ["vellum_custom_nodes", "addition_node"]
    assert node["name"] == "MyAdditionNode"
    assert node["exec_config"] == {
        "adornments": None,
        "attributes": [
            {
                "id": "aed3bcbb-d243-4a77-bb5e-409e9a28e868",
                "name": "arg1",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
            },
            {
                "id": "9225d225-a41b-4642-8964-f28f58dcf4bf",
                "name": "arg2",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
            },
        ],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "addition_node"], "name": "MyAdditionNode"},
        "display_data": {
            "comment": {"value": "Custom node that performs simple addition."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "195cd69d-3d2d-41e4-a432-16c433cb8d34",
        "label": "MyAdditionNode",
        "outputs": [{"id": "3d8e40cb-2aa8-44bd-ae6a-708a9fbc4779", "name": "result", "type": "NUMBER", "value": None}],
        "ports": [{"id": "9a9e4ef6-febf-4093-a515-217bbb1373db", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "a5298668-d808-4a45-a62e-790943948e8a", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }



import logging
import math
import multiprocessing
import time
from typing import Optional

from workflow_server.config import CONCURRENCY, MEMORY_LIMIT_MB

logger = logging.getLogger(__name__)

WARN_MEMORY_PERCENT = 0.90
FORCE_GC_MEMORY_PERCENT = 0.75

_MAX_PROCESS_COUNT = math.ceil(CONCURRENCY * 1.7)
# Keep these under the 15s knative response start timeout or it may trigger odd behavior
_MEMORY_CHECK_INTERVAL_SECONDS = 2
_MAX_MEMORY_CHECK_ATTEMPTS = 3
_ACTIVE_PROCESS_COUNT = multiprocessing.Value("i", 0)
_ACTIVE_PROCESS_LOCK = multiprocessing.Lock()


def increment_process_count(change: int) -> None:
    result = _ACTIVE_PROCESS_LOCK.acquire(timeout=5)
    try:
        if result:
            global _ACTIVE_PROCESS_COUNT
            _ACTIVE_PROCESS_COUNT.value += change  # type: ignore
        else:
            logger.error("Failed to lock workflow server process count global.")
    finally:
        if result:
            _ACTIVE_PROCESS_LOCK.release()


def get_active_process_count() -> int:
    return _ACTIVE_PROCESS_COUNT.value  # type: ignore


def get_memory_in_use_mb() -> Optional[float]:
    try:
        with open("/sys/fs/cgroup/memory/memory.usage_in_bytes", "r") as file:
            memory_bytes = file.read()
    except Exception:
        logger.error("Unable to get current memory.")
        return None

    if not memory_bytes:
        logger.error("Unable to get current memory.")
        return None

    return int(memory_bytes) / 1024 / 1024


def wait_for_available_process() -> bool:
    memory_loops = 0
    process_available = False

    while memory_loops < _MAX_MEMORY_CHECK_ATTEMPTS:
        memory_mb = get_memory_in_use_mb()

        exceeded_warn_limit = memory_mb and memory_mb > (MEMORY_LIMIT_MB * WARN_MEMORY_PERCENT)
        exceeded_process_limit = (
            get_active_process_count() > _MAX_PROCESS_COUNT
            and memory_mb
            and memory_mb > (MEMORY_LIMIT_MB * FORCE_GC_MEMORY_PERCENT)
        )

        if not exceeded_process_limit and not exceeded_warn_limit:
            process_available = True
            break

        memory_loops += 1
        time.sleep(_MEMORY_CHECK_INTERVAL_SECONDS)

    return process_available



import logging
import os
from typing import Any, Optional

from gunicorn import glogging
import gunicorn.app.base

from workflow_server.config import ENABLE_PROCESS_WRAPPER, PORT, is_development
from workflow_server.server import app
from workflow_server.utils.exit_handler import gunicorn_exit_handler, init_signal_handlers
from workflow_server.utils.oom_killer import start_oom_killer_worker


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, app_param: gunicorn.app.base.BaseApplication, options: Optional[dict] = None) -> None:
        self.options = options or {}
        self.application = app_param
        super().__init__()

    def load_config(self) -> None:
        config = {key: value for key, value in self.options.items() if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> None:
        return self.application


class CustomGunicornLogger(glogging.Logger):
    def setup(self, cfg: dict) -> None:
        super().setup(cfg)

        logger = logging.getLogger("gunicorn.access")
        logger.addFilter(HealthCheckFilter())


class HealthCheckFilter(logging.Filter):
    def filter(self, record: Any) -> bool:
        return "GET /healthz" not in record.getMessage()


def start() -> None:
    if not is_development():
        start_oom_killer_worker()
    init_signal_handlers()

    # gevent didn't play nice with pebble and processes so just gthread here
    options = {
        "bind": f"0.0.0.0:{PORT}",
        "workers": int(os.getenv("GUNICORN_WORKERS", 2)),
        "threads": int(os.getenv("GUNICORN_THREADS", 9 if ENABLE_PROCESS_WRAPPER else 6)),
        # Try to avoid memory leaks when using non process mode
        "max_requests": 0 if ENABLE_PROCESS_WRAPPER else 20,
        "worker_class": "gthread",
        "timeout": int(os.getenv("GUNICORN_TIMEOUT", 1800)),
        "logger_class": CustomGunicornLogger,
        "accesslog": "-",
        "on_exit": gunicorn_exit_handler,
        "graceful_timeout": int(os.getenv("GUNICORN_TIMEOUT", 1800)),
    }
    StandaloneApplication(app, options).run()



import gc
import logging
import multiprocessing
from multiprocessing.synchronize import Event
import os
import signal
import sys
from threading import Thread
import time
from time import sleep

from workflow_server.config import MEMORY_LIMIT_MB
from workflow_server.utils.exit_handler import process_killed_switch
from workflow_server.utils.system_utils import (
    FORCE_GC_MEMORY_PERCENT,
    WARN_MEMORY_PERCENT,
    get_active_process_count,
    get_memory_in_use_mb,
)

logger = logging.getLogger(__name__)

_oom_killed_switch = multiprocessing.Event()

_MAX_MEMORY_PERCENT = 0.97
_FORCE_COLLECT_MEMORY_PERCENT = 0.90
_KILL_GRACE_PERIOD = 5


def start_oom_killer_worker() -> None:
    logger.info("Starting oom killer watcher...")
    OomKillerThread(kill_switch=_oom_killed_switch).start()


def get_is_oom_killed() -> bool:
    return _oom_killed_switch.is_set()


class OomKillerThread(Thread):
    """
    This worker is for watching for oom errors so we can gracefully kill any workflows in flight
    and tell the user they have a memory limit problem instead of relying on kubernetes to do it
    for us. This currently goes off the max memory not the requested memory so it may not always
    be accurate since max memory may not always be available
    """

    _kill_switch: Event

    def __init__(
        self,
        kill_switch: Event,
    ) -> None:
        Thread.__init__(self)
        self._kill_switch = kill_switch

    def run(self) -> None:
        last_gc = time.time()

        logger.info("Starting oom watcher...")
        if not MEMORY_LIMIT_MB:
            return

        while True:
            if process_killed_switch.is_set():
                exit(1)
            sleep(1)

            memory_mb = get_memory_in_use_mb()
            if not memory_mb:
                return

            if memory_mb > (MEMORY_LIMIT_MB * _MAX_MEMORY_PERCENT):
                self._kill_switch.set()
                logger.error(
                    f"Workflow server OOM killed, memory: {memory_mb}MB, Process Count: {get_active_process_count()}"
                )
                # Give time for the threads to get our kill switch
                sleep(_KILL_GRACE_PERIOD)
                pid = os.getpid()
                os.kill(pid, signal.SIGKILL)
                sys.exit(1)

            if memory_mb > (MEMORY_LIMIT_MB * WARN_MEMORY_PERCENT):
                logger.warning(
                    f"Memory usage exceeded 90% of limit, memory: {memory_mb}MB, "
                    f"Process Count: {get_active_process_count()}"
                )

            if memory_mb > (MEMORY_LIMIT_MB * FORCE_GC_MEMORY_PERCENT):
                if time.time() - last_gc >= 20:
                    logger.info("Forcing garbage collect from memory pressure")
                    gc.collect()



import pytest
from uuid import uuid4

from workflow_server.server import create_app


@pytest.fixture
def mock_sentry_capture_envelope(mocker):
    mock_transport = mocker.patch("sentry_sdk.client.make_transport")
    return mock_transport.return_value.capture_envelope


def test_sentry_integration_with_workflow_endpoints(monkeypatch, mock_sentry_capture_envelope):
    # GIVEN sentry is configured
    monkeypatch.setenv("SENTRY_DSN", "https://test-dsn@sentry.io/1234567890")

    # AND our /workflow/stream endpoint raises an exception
    def mock_get_version():
        raise Exception("Test exception")

    monkeypatch.setattr("workflow_server.api.workflow_view.get_version", mock_get_version)

    # AND we have a mock trace_id
    trace_id = str(uuid4())

    # AND we have a mock request body
    body = {
        "execution_id": uuid4(),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    pass
""",
        },
        "execution_context": {
            "trace_id": trace_id,
            "parent_context": {
                "type": "API_REQUEST",
                "span_id": str(uuid4()),
                "parent": None,
            },
        },
    }

    # WHEN we call the /workflow/version endpoint
    flask_app = create_app()

    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=body)

        # THEN we get a 500 error
        assert response.status_code == 500

        # AND sentry captures the error with the correct data
        assert mock_sentry_capture_envelope.call_count == 1
        envelope = mock_sentry_capture_envelope.call_args[0][0]
        event = envelope.get_event()
        assert event["level"] == "error"
        assert "Test exception" in event["exception"]["values"][0]["value"]

        # AND the trace_id is tagged
        assert event["tags"]["vellum_trace_id"] == trace_id



from dataclasses import field
from uuid import UUID
from typing import Any, Optional

from vellum.client.core import UniversalBaseModel
from vellum.workflows.context import ExecutionContext

DEFAULT_TIMEOUT_SECONDS = 60 * 30


class BaseExecutorContext(UniversalBaseModel):
    inputs: dict
    state: Optional[dict] = None
    timeout: int = DEFAULT_TIMEOUT_SECONDS
    files: dict[str, str]
    environment_api_key: str
    execution_id: UUID
    module: str
    execution_context: ExecutionContext = field(default_factory=ExecutionContext)
    request_start_time: int
    stream_start_time: int = 0
    vembda_public_url: Optional[str] = None
    node_output_mocks: Optional[list[Any]] = None
    environment_variables: Optional[dict[str, str]] = None

    @property
    def container_overhead_latency(self) -> int:
        return self.stream_start_time - self.request_start_time if self.stream_start_time else -1

    @property
    def trace_id(self) -> UUID:
        return self.execution_context.trace_id

    def __hash__(self) -> int:
        # do we think we need anything else for a unique hash for caching?
        return hash(str(self.execution_id))


class WorkflowExecutorContext(BaseExecutorContext):
    node_id: Optional[UUID] = None  # Sent during run from node UX


class NodeExecutorContext(BaseExecutorContext):
    node_module: str
    node_name: str
